import typing

from dataclasses import dataclass
from mypy.nodes import Argument, ClassDef, Expression, FuncDef, SymbolTable, TypeInfo
from mypy.plugin import ClassDefContext, Plugin, SemanticAnalyzerPluginInterface
from mypy.semanal import SemanticAnalyzer
from mypy.types import Instance, TypeType
from mypy.types import Type as MypyType
from mypy.types import TypeVarType
from typed_ray import build_ray_types, utils


class ActorClassTransformer:
    """Transforms Actor class to a new class with a new base class."""

    def __init__(self, ctx: ClassDefContext):
        self.ctx = ctx
        self.api: SemanticAnalyzer = ctx.api
        self.class_def: ClassDef = ctx.cls
        self.reason: Expression = ctx.reason
        self.info: TypeInfo = ctx.cls.info
        self.symbol_table: SymbolTable = self.info.names
        # TODO: Figure out why pyright complains about this.
        self.body: typing.List[FuncDef] = self.class_def.defs.body  # type: ignore

    def transform(self) -> None:
        """Transforms Actor class to a new class with a new base class."""
        self._add_remote_init_method()
        self._add_options_method()

    def _add_remote_init_method(self) -> None:
        """Add remote method to the class.

        @ray.remote
        class A:
            def __init__(self):
                pass

        a = A.remote()

        This method lets mypy return type of a as ActorHandle[A].
        """
        self._add_class_method(
            name="remote",
            args=self._get_method_args("__init__"),
            return_type=build_ray_types.build_actor_class_instance(
                ctx=self.ctx,
                methods=self._get_ray_exposed_methods(),
                decorated_cls=self._get_my_type(),
            ),
        )

    def _add_options_method(self) -> None:
        """Add options method to the class.

        @ray.remote
        class A:
            def __init__(self):
                pass

        a = A.options(...).remote()
        """
        self._add_class_method(
            name="options",
            args=build_ray_types.build_options_args(self.api),
            return_type=TypeType(self._get_my_type()),
        )

    def _add_class_method(
        self,
        name: str,
        args: typing.List[Argument],
        return_type: MypyType,
        tvar_def: typing.Optional[TypeVarType] = None,
    ):
        # First remove any previously generated methods with the same name
        # to avoid clashes and problems in the semantic analyzer.
        if name in self.info.names:
            sym = self.info.names[name]
            if sym.plugin_generated and isinstance(sym.node, FuncDef):  # type: ignore
                self.class_def.defs.body.remove(sym.node)  # type: ignore

        assert isinstance(self.api, SemanticAnalyzerPluginInterface)
        function_type = self.api.named_type("builtins.function")

        func = utils.build_func_def(
            args=args,
            return_type=return_type,
            fallback=function_type,
            func_info=self.info,
            name=name,
            tvar_def=tvar_def,
            is_static=True,
        )
        utils.add_method(name=name, func=func, cls_info=self.info)

    def _get_method_args(self, method_name: str) -> typing.List[Argument]:
        """Get arguments of method belonging to the class.

        class A:
            def foo(a: int, b: Dict[str, int]):
                pass

        _get_method_args("__init__")  # Returns [Argument(a, ...), Argument(b, ...)]
        or rather their mypy representations.
        """
        args = []
        for stmt in self.body:
            if isinstance(stmt, FuncDef) and stmt.name == method_name:
                # Skip self argument.
                args = stmt.arguments[1:]  # type: ignore
                break
        return args  # type: ignore

    def _get_ray_exposed_methods(self) -> typing.List[FuncDef]:
        """Get methods of the decorated class."""
        return [
            stmt
            for stmt in self.body
            if isinstance(stmt, FuncDef) and not utils.is_magic_method(stmt.name)
        ]

    def _get_my_type(self) -> MypyType:
        """Get mypy representation of the class."""
        return Instance(self.info, [])


@dataclass()
class RayWorkerDecoratorCallback:
    plugin: Plugin
    fullname: str

    def __call__(self, ctx: ClassDefContext) -> None:
        transformer = ActorClassTransformer(ctx)
        transformer.transform()



from mypy.nodes import ClassDef as ClassDef
from mypy.plugin import ClassDefContext as ClassDefContext, Plugin as Plugin
from typing import Any

class ActorClassTransformer:
    ctx: Any
    api: Any
    class_def: Any
    reason: Any
    info: Any
    symbol_table: Any
    body: Any
    def __init__(self, ctx: ClassDefContext) -> None: ...
    def transform(self) -> ClassDef: ...

class RayWorkerDecoratorCallback:
    plugin: Plugin
    fullname: str
    def __call__(self, ctx: ClassDefContext) -> None: ...
    def __init__(self, plugin: Plugin, fullname: str) -> None: ...



# pyright: reportPrivateUsage=false
import typing
from pprint import pformat, pprint

from mypy.nodes import (
    ARG_POS,
    MDEF,
    ArgKind,
    Argument,
    Block,
    ClassDef,
    FuncDef,
    PassStmt,
    SymbolTableNode,
    TypeInfo,
    TypeVarExpr,
    Var,
)
from mypy.plugin import ClassDefContext, DynamicClassDefContext
from mypy.semanal import SemanticAnalyzer
from mypy.semanal_shared import set_callable_name
from mypy.types import AnyType, CallableType, Instance, NoneType
from mypy.types import Type as MypyType
from mypy.types import TypeOfAny, TypeVarType, UnboundType, UnionType
from mypy.typevars import fill_typevars


def build_argument(
    name: str,
    type_: typing.Optional[MypyType] = None,
    arg_type: ArgKind = ARG_POS,
) -> Argument:
    """Builds a function argument."""
    if type_ is None:
        type_ = AnyType(TypeOfAny.unannotated)
    return Argument(Var(name, type_), type_, None, arg_type)


def build_func_def(
    args: typing.List[Argument],
    return_type: MypyType,
    fallback: Instance,
    func_info: TypeInfo,
    name: str,
    tvar_def: typing.Optional[TypeVarType] = None,
    self_type: typing.Optional[MypyType] = None,
    is_static: bool = False,
    is_property: bool = False,
    body: typing.Optional[Block] = None,
) -> FuncDef:
    if not is_static:
        self_type = self_type or fill_typevars(func_info)
        args = [Argument(Var("self"), self_type, None, ARG_POS)] + args
    arg_types, arg_names, arg_kinds = [], [], []
    for arg in args:
        assert arg.type_annotation, "All arguments must be fully typed."  # type: ignore
        arg_types.append(arg.type_annotation)  # type: ignore
        arg_names.append(arg.variable.name)  # type: ignore
        arg_kinds.append(arg.kind)  # type: ignore

    signature = CallableType(arg_types, arg_kinds, arg_names, return_type, fallback)  # type: ignore
    if tvar_def:
        signature.variables = [tvar_def]
    if not body:
        body = Block([PassStmt()])
    func = FuncDef(name, args, body)
    func.info = func_info
    func.type = set_callable_name(signature, func)
    func._fullname = func_info.fullname + "." + name  # type: ignore
    func.line = func_info.line  # type: ignore
    func.is_static = is_static
    func.is_property = is_property
    return func


def add_method(
    name: str, func: FuncDef, cls_info: TypeInfo, replace: bool = False
) -> None:
    # First remove any previously generated methods with the same name
    # to avoid clashes and problems in the semantic analyzer.
    if name in cls_info.names:
        sym = cls_info.names[name]
        if sym.plugin_generated and isinstance(sym.node, FuncDef):  # type: ignore
            cls_info.defn.defs.body.remove(sym.node)  # type: ignore
    # NOTE: we would like the plugin generated node to dominate, but we still
    # need to keep any existing definitions so they get semantically analyzed.
    if name in cls_info.names and not replace:
        # Get a nice unique name instead.
        return
        # r_name = get_unique_redefinition_name(name, cls_info.names)
        # cls_info.names[r_name] = cls_info.names[name]

    cls_info.names[name] = SymbolTableNode(MDEF, func, plugin_generated=True)
    cls_info.defn.defs.body.append(func)  # type: ignore


def add_attribute(name: str, cls: ClassDef, type_: MypyType) -> None:
    var = Var(name)
    var.info = cls.info
    var.type = type_
    var._fullname = cls.info.fullname + "." + name  # type: ignore
    cls.info.names[name] = SymbolTableNode(MDEF, var)


def add_global(
    ctx: typing.Union[ClassDefContext, DynamicClassDefContext],
    module: str,
    symbol_name: str,
    asname: str,
) -> None:
    module_globals = ctx.api.modules[ctx.api.cur_mod_id].names
    if asname not in module_globals:
        lookup_sym: SymbolTableNode = ctx.api.modules[module].names[symbol_name]

        module_globals[asname] = lookup_sym


def add_remote_method_to_context(ctx: ClassDefContext) -> None:
    add_global(
        ctx,
        "typed_ray.ray_types",
        "RemoteFunction",
        "__tr_RemoteMethod",
    )


def add_actor_class_to_context(ctx: ClassDefContext) -> None:
    add_global(
        ctx,
        "typed_ray.ray_types",
        "ActorClass",
        "__tr_ActorClass",
    )


def add_object_ref_to_context(ctx: ClassDefContext) -> None:
    add_global(
        ctx,
        "typed_ray.ray_types",
        "ObjectRef",
        "__tr_ObjectRef",
    )


def is_magic_method(name: str) -> bool:
    return name.startswith("__") and name.endswith("__")


def get_return_type(method: FuncDef) -> MypyType:
    try:
        return method.type.ret_type  # type: ignore
    except AttributeError:
        return AnyType(TypeOfAny.unannotated)


def var_to_type(var: Var) -> MypyType:
    return var.type or AnyType(TypeOfAny.unannotated)


def unbound_to_instance(api: SemanticAnalyzer, type_: MypyType) -> MypyType:
    if not isinstance(type_, UnboundType):
        return type_

    type_name: str = type_.name  # type: ignore

    if type_name == "Optional":
        # convert from "Optional?" to the more familiar
        # UnionType[..., NoneType()]
        return unbound_to_instance(
            api,
            UnionType(
                [unbound_to_instance(api, typ_arg) for typ_arg in type_.args]  # type: ignore
                + [NoneType()]
            ),
        )

    node = api.lookup(type_name, type_)

    if node is None or not isinstance(node, SymbolTableNode):  # type: ignore
        return type_
    bound_type = node.node  # type: ignore
    if isinstance(bound_type, Var) and bound_type.name == "None":
        return NoneType()
    if isinstance(bound_type, Var):
        bound_type = var_to_type(bound_type)  # type: ignore
    args: typing.List[MypyType] = []
    for arg in type_.args:  # type: ignore
        if isinstance(arg, UnboundType):
            args.append(unbound_to_instance(api, arg))
        elif isinstance(arg, TypeVarExpr):
            args.extend(arg.values)
        else:
            assert isinstance(arg, MypyType)
            args.append(arg)
    assert isinstance(bound_type, TypeInfo)
    return Instance(bound_type, args)  # type: ignore


def pformat_(x: typing.Any, seen: typing.Optional[typing.Set[int]]=None) -> typing.Any:
    if seen is None:
        seen = set()
    if isinstance(x, list):
        if not x:
            return "[]"
        list_result: typing.List[str] = []
        for y in x:  # type: ignore
            list_result.append(pformat_(y, seen))
            seen.add(id(y))  # type: ignore
        return list_result
    if isinstance(x, str):
        return x
    result: typing.Dict[str, typing.Any] = {}
    for name in dir(x):
        try:
            value = getattr(x, name)
        except (AttributeError, AssertionError):
            continue
        value_id = id(value)
        if name.startswith("__") and name.endswith("__"):
            continue
        if value_id in seen:
            continue
        seen.add(value_id)
        result[name] = pformat_(value, seen)
    result["class"] = type(x)
    result["id"] = id(x)
    return result


def print_(x: typing.Any) -> None:
    as_str = f"----------{type(x)}----------\n"
    as_str += pformat(pformat_(x))
    as_str += "\n"
    as_str += "-----------------------------"
    pprint(as_str)



from pathlib import Path

import toml

PYPROJECT_PATH = Path(__file__).parent.parent / 'pyproject.toml'

def parse_version_from_pyproject_file(pyproject_file: Path) -> str:
    """
    Parse the version from a pyproject.toml file.
    """
    with open(pyproject_file) as f:
        pyproject_data = toml.load(f)
    return pyproject_data['tool']['poetry']['version']


__version__ = parse_version_from_pyproject_file(PYPROJECT_PATH)



import typing
from mypy.nodes import (
    ArgKind,
    Argument as Argument,
    Block as Block,
    ClassDef as ClassDef,
    FuncDef as FuncDef,
    TypeInfo as TypeInfo,
    Var as Var,
)
from mypy.plugin import (
    ClassDefContext as ClassDefContext,
    DynamicClassDefContext as DynamicClassDefContext,
)
from mypy.semanal import SemanticAnalyzer as SemanticAnalyzer
from mypy.types import (
    Instance as Instance,
    Type as MypyType,
    TypeVarType as TypeVarType,
    UnboundType as UnboundType,
)
from typing import Any

def build_argument(
    name: str,
    type_: typing.Optional[MypyType] = ...,
    arg_type: ArgKind = ...
) -> Argument: ...
def build_func_def(
    args: typing.List[Argument],
    return_type: MypyType,
    fallback: Instance,
    func_info: TypeInfo,
    name: str,
    tvar_def: typing.Optional[TypeVarType] = ...,
    self_type: MypyType = ...,
    is_static: bool = ...,
    is_property: bool = ...,
    body: Block = ...,
) -> FuncDef: ...
def add_method(
    name: str, func: FuncDef, cls_info: TypeInfo, replace: bool = ...
) -> None: ...
def add_attribute(name: str, cls: ClassDef, type_: MypyType) -> None: ...
def add_global(
    ctx: typing.Union[ClassDefContext, DynamicClassDefContext],
    module: str,
    symbol_name: str,
    asname: str,
) -> None: ...
def add_remote_method_to_context(ctx: ClassDefContext) -> None: ...
def add_actor_class_to_context(ctx: ClassDefContext) -> None: ...
def add_object_ref_to_context(ctx: ClassDefContext) -> None: ...
def is_magic_method(name: str) -> bool: ...
def get_return_type(method: FuncDef) -> MypyType: ...
def var_to_type(var: Var) -> MypyType: ...
def unbound_to_instance(api: SemanticAnalyzer, type_: MypyType) -> MypyType: ...
def pformat_(x: typing.Any, seen: typing.Union[Any, None] = ...) -> typing.Any: ...
def print_(x: typing.Any) -> None: ...



import typing
from mypy.plugin import Plugin
from typed_ray import mypy_types as mypy_types

class _RayPlugin(Plugin):
    def get_class_decorator_hook(
        self, fullname: str
    ) -> typing.Optional[mypy_types.ClassDecoratorHookCallback]: ...

def plugin(version: str) -> Plugin: ...



"""Build the types in ray_types using Mypy."""
import typing

from mypy.nodes import ARG_OPT, Argument, Block, ClassDef, FuncDef, SymbolTable, TypeInfo
from mypy.plugin import ClassDefContext
from mypy.semanal import SemanticAnalyzer
from mypy.types import Instance
from mypy.types import Type as MypyType
from mypy.types import TypeVarLikeType
from typed_ray import ray_types, utils


def build_options_args(
    api: SemanticAnalyzer
) -> typing.List[Argument]:
    """Build the arguments for the ActorHandle.options method."""
    return [
        utils.build_argument(
            arg_name, api.named_type(arg_type_str), arg_type=ARG_OPT
        )
        for arg_name, arg_type_str in ray_types.OPTIONS_KWARGS.items()
    ]


def build_type_info(
    api: SemanticAnalyzer,
    module_name: str,
    cls_full_name: str,
    column: int = -1,
    line: int = -1,
    class_def: typing.Optional[ClassDef] = None,
    class_name: typing.Optional[str] = None,
    type_vars: typing.Optional[typing.List[TypeVarLikeType]] = None,
) -> TypeInfo:
    """Build a well constructed TypeInfo object."""
    if type_vars is None:
        type_vars = []
    if class_def is None:
        assert (
            class_name is not None
        ), "One of class_def and class_name must be specified"
        class_def = ClassDef(
            name=class_name,
            defs=Block([]),
            type_vars=type_vars,
        )
        class_def.column = column
        class_def.line = line
        class_def.fullname = cls_full_name
    result = TypeInfo(
        names=SymbolTable(),
        defn=class_def,
        module_name=module_name,
    )
    class_def.info = result
    result._fullname = cls_full_name  # type: ignore
    result.bases = [api.named_type("builtins.object")]
    object_type = api.named_type("builtins.object").type  # type: ignore
    assert isinstance(object_type, TypeInfo)
    result.mro = [result, object_type]
    return result


def build_object_ref(ctx: ClassDefContext, wraps: MypyType) -> MypyType:
    """Build the type of the object reference."""
    utils.add_object_ref_to_context(ctx)
    api: SemanticAnalyzer = ctx.api
    obj_ref_info: TypeInfo = api.lookup_fully_qualified(
        ray_types.ObjectRefFullName
    ).node  # type: ignore
    return Instance(obj_ref_info, [wraps])


def build_remote_method_info(
    api: SemanticAnalyzer,
    module_name: str,
    column: int,
    line: int,
    cls_full_name: str,
) -> TypeInfo:
    return build_type_info(
        api=api,
        module_name=module_name,
        cls_full_name=cls_full_name,
        column=column,
        line=line,
        class_name="RemoteFunction",
    )


def build_remote_method(
    ctx: ClassDefContext,
    method: FuncDef,
) -> MypyType:
    api: SemanticAnalyzer = ctx.api
    info: TypeInfo = method.info  # type: ignore
    remote_method_info = build_remote_method_info(
        api=api,
        module_name=info.module_name,
        cls_full_name=info._fullname,  # type: ignore
        column=method.column,  # type: ignore
        line=method.line,  # type: ignore
    )
    method_ret_type = utils.get_return_type(method)
    bound_method_ret_type = utils.unbound_to_instance(api, method_ret_type)
    return_type = build_object_ref(ctx, bound_method_ret_type)
    remote_method = utils.build_func_def(
        args=method.arguments[1:],  # type: ignore
        return_type=return_type,
        fallback=api.named_type("builtins.function"),
        func_info=remote_method_info,
        name="remote",
        body=method.body,
    )
    utils.add_method(
        name="remote", func=remote_method, cls_info=remote_method_info, replace=True
    )
    return Instance(remote_method_info, [bound_method_ret_type])


def build_actor_class_info(
    ctx: ClassDefContext,
    methods: typing.List[FuncDef],
) -> TypeInfo:
    """Build the Actor class TypeInfo for an object having methods `methods`."""
    api: SemanticAnalyzer = ctx.api
    result = build_type_info(
        api=api,
        module_name="typed_ray.ray_types",
        cls_full_name="typed_ray.ray_types.ActorHandle",
        class_name="ActorHandle",
    )
    for method in methods:
        type_ = build_remote_method(ctx, method)
        utils.add_attribute(name=method.name, cls=result.defn, type_=type_)
    return result


def build_actor_class_instance(
    ctx: ClassDefContext,
    methods: typing.List[FuncDef],
    decorated_cls: MypyType,
) -> MypyType:
    """Build the Actor class instance for an object having methods `methods`."""
    return Instance(build_actor_class_info(ctx=ctx, methods=methods), [decorated_cls])



from ray import (
    init as init,
    is_initialized as is_initialized,
    cross_language as cross_language,
)
from ray.util import ActorPool as ActorPool
from ray.util.queue import Queue as Queue
from typed_ray.ray_types import ActorHandle as ActorHandle, ObjectRef as ObjectRef
from typed_ray.typed_ray import (
    get as get,
    put as put,
    remote_cls as remote_cls,
    remote_func as remote_func,
)



# pyright: reportWildcardImportFromLibrary=false
from ray import (
    init as init,
    is_initialized as is_initialized,
    cross_language as cross_language,
)
from ray.util import ActorPool as ActorPool
from ray.util.queue import Queue as Queue

from typed_ray.typed_ray import *  # noqa: F401
from typed_ray.ray_types import *  # noqa: F401
from typed_ray import _version

__version__ = _version.__version__



import typing

from typing_extensions import ParamSpec

ObjectRefFullName: str

_T = typing.TypeVar("_T")

class ObjectRef(typing.Generic[_T]): ...
class ActorHandle(typing.Generic[_T]): ...

_ArgsT = ParamSpec("_ArgsT")
_ReturnT = typing.TypeVar("_ReturnT")

OPTIONS_KWARGS: typing.Dict[str, str]

class _RemoteFunctionOptions(typing.Protocol[_ArgsT, _ReturnT]):
    def __call__(
        self,
        num_gpus: typing.Optional[int] = ...,
        num_cpus: typing.Optional[int] = ...,
        name: str = ...,
    ) -> "RemoteFunction[_ArgsT, _ReturnT]": ...

class RemoteFunction(typing.Generic[_ArgsT, _ReturnT]):
    @property
    def options(self) -> _RemoteFunctionOptions[_ArgsT, _ReturnT]: ...
    @property
    def remote(self) -> typing.Callable[_ArgsT, ObjectRef[_ReturnT]]: ...



import typing
from mypy.nodes import (
    Argument,
    ClassDef as ClassDef,
    FuncDef as FuncDef,
    SymbolTable as SymbolTable,
    TypeInfo as TypeInfo,
)
from mypy.plugin import ClassDefContext as ClassDefContext
from mypy.semanal import SemanticAnalyzer as SemanticAnalyzer
from mypy.types import Type as MypyType, TypeVarLikeType as TypeVarLikeType

def build_options_args(api: SemanticAnalyzer) -> typing.List[Argument]: ...
def build_type_info(
    api: SemanticAnalyzer,
    module_name: str,
    cls_full_name: str,
    column: int = ...,
    line: int = ...,
    class_def: ClassDef = ...,
    class_name: str = ...,
    type_vars: TypeVarLikeType = ...,
) -> TypeInfo: ...
def build_object_ref(ctx: ClassDefContext, wraps: MypyType) -> MypyType: ...
def build_remote_method_info(
    api: SemanticAnalyzer,
    module_name: str,
    column: int,
    line: int,
    cls_full_name: str = ...,
) -> TypeInfo: ...
def build_remote_method(ctx: ClassDefContext, method: FuncDef) -> MypyType: ...
def build_actor_class_info(
    ctx: ClassDefContext, methods: typing.List[FuncDef]
) -> TypeInfo: ...
def build_actor_class_instance(
    ctx: ClassDefContext, methods: typing.List[FuncDef], decorated_cls: MypyType
) -> MypyType: ...



import typing

import ray
from typing_extensions import ParamSpec

from typed_ray.ray_types import ActorHandle, ObjectRef, RemoteFunction


T = typing.TypeVar("T")


__all__ = [
    "put",
    "get",
    "wait",
    "remote_func",
    "remote_cls",
    "get_actor",
    "kill",
    "cancel",
    "get_gpu_ids",
    "shutdown",
    "method",
    "nodes",
    "cluster_resources",
    "available_resources",
    "timeline",
]


def put(value: T) -> ObjectRef[T]:
    """
    Put a value into the object store.

    Args:
        value: The value to put into the object store.

    Returns:
        An ObjectRef object that can be used to get the value back later.
    """
    return ray.put(value)  # type: ignore


@typing.overload
def get(object_refs: ObjectRef[T]) -> T:
    pass


@typing.overload
def get(object_refs: typing.List[ObjectRef[T]]) -> typing.List[T]:
    pass


def get(
    object_refs: typing.Union[ObjectRef[T], typing.List[ObjectRef[T]]]
) -> typing.Union[T, typing.List[T]]:
    """
    Get the value wrapped by ObjectRef.

    Args:
        object_refs: The ObjectRef object usually returned by ray.put.

    Returns:
        The value that was wrapped by ObjectRef.
    """
    return ray.get(object_refs)  # type: ignore


def wait(
    object_refs: typing.List[ObjectRef[T]],
    *,
    num_returns: int = 1,
    timeout: typing.Optional[float],
    fetch_local: bool = True
) -> typing.Tuple[typing.List[ObjectRef[T]], typing.List[ObjectRef[T]]]:
    """Return a list of IDs that are ready and a list of IDs that are not.

    If timeout is set, the function returns either when the requested number of
    IDs are ready or when the timeout is reached, whichever occurs first. If it
    is not set, the function simply waits until that number of objects is ready
    and returns that exact number of object refs.

    This method returns two lists. The first list consists of object refs that
    correspond to objects that are available in the object store. The second
    list corresponds to the rest of the object refs (which may or may not be
    ready).

    Ordering of the input list of object refs is preserved. That is, if A
    precedes B in the input list, and both are in the ready list, then A will
    precede B in the ready list. This also holds true if A and B are both in
    the remaining list.

    This method will issue a warning if it's running inside an async context.
    Instead of ``ray.wait(object_refs)``, you can use
    ``await asyncio.wait(object_refs)``.

    Args:
        object_refs (List[ObjectRef]): List of object refs for objects that may
            or may not be ready. Note that these IDs must be unique.
        num_returns (int): The number of object refs that should be returned.
        timeout (float): The maximum amount of time in seconds to wait before
            returning.
        fetch_local (bool): If True, wait for the object to be downloaded onto
            the local node before returning it as ready. If False, ray.wait()
            will not trigger fetching of objects to the local node and will
            return immediately once the object is available anywhere in the
            cluster.

    Returns:
        A list of object refs that are ready and a list of the remaining object
        IDs.
    """
    return ray.wait(  # type: ignore
        object_refs, num_returns=num_returns, timeout=timeout, fetch_local=fetch_local
    )


_ArgsT = ParamSpec("_ArgsT")
_ReturnT = typing.TypeVar("_ReturnT")


def remote_func(
    func: typing.Callable[_ArgsT, _ReturnT]
) -> RemoteFunction[_ArgsT, _ReturnT]:
    """
    Make a remote function.

    Args:
        func: The function to make remote.

    Returns:
        A remote function.
    """
    return ray.remote(func)  # type: ignore


def remote_cls(cls: typing.Type[typing.Any]) -> typing.Any:
    """
    Make a remote class.

    Args:
        cls: The class to make remote.

    Returns:
        A remote class.
    """
    return ray.remote(cls)  # type: ignore


def get_actor(
    name: str, namespace: typing.Optional[str] = None
) -> ActorHandle[typing.Any]:
    """Get a handle to a namd actor.

    Gets a handle to an actor with the given name. The actor must
    have been created with Actor.options(name="name").remote(). This
    works for both detached & non-detached actors.

    Args:
        name: The name of the actor.
        namespace: The namespace of the actor, or None to specify the current
            namespace.

    Returns:
        ActorHandle to the actor.

    Raises:
        ValueError if the named actor does not exist.
    """
    return ray.get_actor(name, namespace)  # type: ignore


def kill(actor: ActorHandle[typing.Any]) -> None:
    """Kill an actor forcefully.

    This will interrupt any running tasks on the actor, causing them to fail
    immediately. ``atexit`` handlers installed in the actor will not be run.

    If you want to kill the actor but let pending tasks finish,
    you can call ``actor.__ray_terminate__.remote()`` instead to queue a
    termination task. Any ``atexit`` handlers installed in the actor *will*
    be run in this case.

    If the actor is a detached actor, subsequent calls to get its handle via
    ray.get_actor will fail.

    Args:
        actor (ActorHandle): Handle to the actor to kill.
        no_restart (bool): Whether or not this actor should be restarted if
            it's a restartable actor.
    """
    ray.kill(actor)  # type: ignore


def cancel(
    object_ref: ObjectRef[typing.Any], *, force: bool = False, recursive: bool = True
) -> None:
    """Cancels a task according to the following conditions.

    If the specified task is pending execution, it will not be executed. If
    the task is currently executing, the behavior depends on the ``force``
    flag. When ``force=False``, a KeyboardInterrupt will be raised in Python
    and when ``force=True``, the executing task will immediately exit.
    If the task is already finished, nothing will happen.

    Only non-actor tasks can be canceled. Canceled tasks will not be
    retried (max_retries will not be respected).

    Calling ray.get on a canceled task will raise a TaskCancelledError or a
    WorkerCrashedError if ``force=True``.

    Args:
        object_ref (ObjectRef): ObjectRef returned by the task
            that should be canceled.
        force (boolean): Whether to force-kill a running task by killing
            the worker that is running the task.
        recursive (boolean): Whether to try to cancel tasks submitted by the
            task specified.
    Raises:
        TypeError: This is also raised for actor tasks.
    """
    return ray.kill(object_ref, force=force, recursive=recursive)  # type: ignore


def get_gpu_ids() -> typing.List[str]:
    """Get the IDs of the GPUs that are available to the worker.

    If the CUDA_VISIBLE_DEVICES environment variable was set when the worker
    started up, then the IDs returned by this method will be a subset of the
    IDs in CUDA_VISIBLE_DEVICES. If not, the IDs will fall in the range
    [0, NUM_GPUS - 1], where NUM_GPUS is the number of GPUs that the node has.

    Returns:
        A list of GPU IDs.
    """
    return ray.get_gpu_ids()  # type: ignore


def shutdown(_exiting_interpreter: bool = False) -> None:
    """Disconnect the worker, and terminate processes started by ray.init().

    This will automatically run at the end when a Python process that uses Ray
    exits. It is ok to run this twice in a row. The primary use case for this
    function is to cleanup state between tests.

    Note that this will clear any remote function definitions, actor
    definitions, and existing actors, so if you wish to use any previously
    defined remote functions or actors after calling ray.shutdown(), then you
    need to redefine them. If they were defined in an imported module, then you
    will need to reload the module.

    Args:
        _exiting_interpreter (bool): True if this is called by the atexit hook
            and false otherwise. If we are exiting the interpreter, we will
            wait a little while to print any extra error messages.
    """
    return ray.shutdown()  # type: ignore


def method(num_returns: int) -> typing.Any:
    """Annotate an actor method.

    .. code-block:: python

        @ray.remote
        class Foo:
            @ray.method(num_returns=2)
            def bar(self):
                return 1, 2

        f = Foo.remote()

        _, _ = f.bar.remote()

    Args:
        num_returns: The number of object refs that should be returned by
            invocations of this actor method.
    """
    return ray.method(num_returns)  # type: ignore


def nodes() -> typing.List[
    typing.Dict[str, typing.Union[float, typing.Dict[str, float]]]
]:
    """Get a list of the nodes in the cluster (for debugging only).

    Returns:
        Information about the Ray clients in the cluster.
    """
    return ray.nodes()  # type: ignore


def cluster_resources() -> typing.Dict[str, float]:
    """Get the current total cluster resources.

    Note that this information can grow stale as nodes are added to or removed
    from the cluster.

    Returns:
        A dictionary mapping resource name to the total quantity of that
            resource in the cluster.
    """
    return ray.cluster_resources()  # type: ignore


def available_resources() -> typing.Dict[str, float]:
    """Get the current available cluster resources.

    This is different from `cluster_resources` in that this will return idle
    (available) resources rather than total resources.

    Note that this information can grow stale as tasks start and finish.

    Returns:
        A dictionary mapping resource name to the total quantity of that
            resource in the cluster.
    """
    return ray.available_resources()  # type: ignore


def timeline() -> typing.List[typing.Dict[str, typing.Any]]:
    """Return a list of profiling events that can viewed as a timeline.

    Ray profiling must be enabled by setting the RAY_PROFILING=1 environment
    variable prior to starting Ray.

    To view this information as a timeline, simply dump it as a json file by
    passing in "filename" or using using json.dump, and then load go to
    chrome://tracing in the Chrome web browser and load the dumped file.

    Args:
        filename: If a filename is provided, the timeline is dumped to that
            file.

    Returns:
        If filename is not provided, this returns a list of profiling events.
            Each profile event is a dictionary.
    """
    return ray.timeline()  # type: ignore



import typing
from typing_extensions import ParamSpec

_T = typing.TypeVar("_T")
_ArgsT = ParamSpec("_ArgsT")
_ReturnT = typing.TypeVar("_ReturnT")

ObjectRefFullName = "typed_ray.ray_types.ObjectRef"


__all__ = ["OPTIONS_KWARGS", "ObjectRef", "ActorHandle"]


OPTIONS_KWARGS = {
    "num_gpus": "builtins.int", "num_cpus": "builtins.int", "name": "builtins.str"
}


class ObjectRef(typing.Generic[_T]):
    """Object returned by ray.put"""


class ActorHandle(typing.Generic[_T]):
    """Actor handle returned by decorating class with tray.remote_cls"""


class _RemoteFunctionOptions(typing.Protocol[_ArgsT, _ReturnT]):
    """Options for remote function"""
    def __call__(
        self,
        num_gpus: typing.Optional[int] = None,
        num_cpus: typing.Optional[int] = None,
        name: str = "",
    ) -> "RemoteFunction[_ArgsT, _ReturnT]": ...



class RemoteFunction(typing.Generic[_ArgsT, _ReturnT]):
    @property
    def remote(self) -> typing.Callable[_ArgsT, ObjectRef[_ReturnT]]:
        """
        Call a remote function.
        """
        raise NotImplementedError

    @property
    def options(self) -> _RemoteFunctionOptions[_ArgsT, _ReturnT]:
        """
        Set options for a remote function.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['typed_ray', 'typed_ray.callbacks']

package_data = \
{'': ['*']}

install_requires = \
['mypy>=0.931,<0.932', 'ray>=1.9.1,<2.0.0']

setup_kwargs = {
    'name': 'typed-ray',
    'version': '0.2.1',
    'description': 'Mypy plugin to add support for type checking ray objects.',
    'long_description': None,
    'author': 'Adi Gudimella',
    'author_email': 'aditya.gudimella@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



import typing

from mypy.plugin import Plugin
from typed_ray import callbacks, constants, mypy_types


class _RayPlugin(Plugin):
    _class_decorator_hook_plugins: typing.Dict[
        str, typing.Callable[[Plugin, str], mypy_types.ClassDecoratorHookCallback]
    ] = {constants.TYPED_RAY_REMOTE_DECORATOR: callbacks.RayWorkerDecoratorCallback}

    def get_class_decorator_hook(
        self, fullname: str
    ) -> typing.Optional[mypy_types.ClassDecoratorHookCallback]:
        if fullname.endswith("typed_ray.remote_cls"):
            return self._class_decorator_hook_plugins[
                constants.TYPED_RAY_REMOTE_DECORATOR
            ](self, fullname)


def plugin(version: str) -> typing.Type[Plugin]:
    return _RayPlugin



import typing
from typing_extensions import ParamSpec
from typed_ray.ray_types import ObjectRef as ObjectRef, RemoteFunction as RemoteFunction

T = typing.TypeVar("T")
_ArgsT = ParamSpec("_ArgsT")
_ReturnT = typing.TypeVar("_ReturnT")

def put(value: T) -> ObjectRef[T]: ...
@typing.overload
def get(object_refs: ObjectRef[T]) -> T: ...
@typing.overload
def get(object_refs: typing.List[ObjectRef[T]]) -> typing.List[T]: ...
def remote_func(
    func: typing.Callable[_ArgsT, _ReturnT]
) -> RemoteFunction[_ArgsT, _ReturnT]: ...
def remote_cls(cls: typing.Type[typing.Any]) -> typing.Any: ...
